local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("Bars")

local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0-Plugins")

local LSM = LibStub("LibSharedMedia-3.0")

local defaults = {
	profile = {
		fadeTime = 0.5,
		fadeOutTime = 2,
		Enabled = {
			Long = true,
			Normal = true,
		},
		Master = {
			UseGlobal = true,
			GrowUp = true
		},
		Animate = true,
		AnimSpeed = 2,
		Anchors = {
			Short = {
				x = 0,
				y = -100,
				UseGlobal = false,
				UseMaster = true,
				Size = {Width = 320, Height = 46, Spacing = 0},
				Fill = true,
				GrowUp = true,
				Time = 12,
				Enabled = true,
				HavePos = false,
			},
			Long = {
				x = -600,
				y = 100,
				UseGlobal = false,
				UseMaster = true,
				Size = {Width = 280, Height = 32, Spacing = 0},
				Fill = true,
				GrowUp = true,
				Time = 1000000,
				Enabled = true,
				HavePos = false,
			},
			Normal = {
				x = -600,
				y = -100,
				UseGlobal = false,
				UseMaster = true,
				Size = {Width = 280, Height = 32, Spacing = 0},
				Fill = true,
				GrowUp = true,
				Enabled = true,
				Time = 70,
				HavePos = false,
			},
		}
	}
}

for i, setting in pairs(RW.defaults.profile) do
	defaults.profile.Master[i] = setting
end

local inset = defaults.profile.Master.BorderSize > 2 and defaults.profile.Master.Insets or 0
local backdrop = {
	bgFile = LSM:Fetch("background", defaults.profile.Master.BGTexture), 
	edgeFile = LSM:Fetch("border", defaults.profile.Master.Border), 
	tile = false,
	edgeSize = defaults.profile.Master.BorderSize,
	insets = { left = inset, right = inset, top = inset, bottom = inset },
}

function plugin:OnRegister()
	self.db = RW.db:RegisterNamespace(self:GetName(), defaults)
	
	self.freeBarFrames = {}
	self.freeBarObjects = {}
	self.runningBars = {}
end

function plugin:PluginEnable()
	self:RegisterMessage("RW_ConfigMode")
	self:RegisterMessage("RW_DefaultsChanged")
	self:RegisterMessage("RW_StartBar")
	self:RegisterMessage("RW_CancelBar")
	
	self.barFrameCount = 0
end

function plugin:PluginDisable()
end

function plugin:EnableUpdate(enable)
	local function onUpdate(self, elapsed)
		for i, bar in pairs(plugin.runningBars) do
			bar:Update(elapsed)
		end
	end
	if not self.updateFrame then self.updateFrame = CreateFrame("Frame", "RWB_UpdateFrame") end
	if enable then
		self.updateFrame:SetScript("OnUpdate", onUpdate)
	else
		self.updateFrame:SetScript("OnUpdate", nil)
	end
end

local function dragStart(self)
	self:StartMoving()
end
local function dragStop(self)
	local db = plugin.db.profile.Anchors[self.anchorName]
	self:StopMovingOrSizing()
	local s = self:GetEffectiveScale()
	db.x = self:GetLeft() * s
	db.y = self:GetBottom() * s
	db.HavePos = true
end

local function createAnchor(name)
	local db = plugin.db.profile.Anchors[name] or plugin.db.profile.Master
	plugin.anchors = plugin.anchors or {}
	local anchor = RW:CreateAnchor(name, UIParent, name.." bars")
	anchor:ClearAllPoints()
	local s = anchor:GetEffectiveScale()
	if db.HavePost then
		anchor:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", db.x / s, db.y / s)
	else		
		anchor:SetPoint("CENTER", UIParent, "CENTER", db.x, db.y)
	end
	anchor:SetWidth(db.Size.Width)
	anchor:SetScript("OnDragStart", dragStart)
	anchor:SetScript("OnDragStop", dragStop)
	anchor.bars = {}
	anchor.anchorName = name
	plugin.anchors[name] = anchor
end

local function createAnchors()
	for i, setting in pairs(plugin.db.profile.Anchors) do
		createAnchor(i)
	end
end

local function rightClick(self, button)
	if button == "RightButton" then
		self.obj:Stop()
	end
end

local function barSort(a, b)
	return a.timeLeft < b.timeLeft
end

local barProt = {}
function barProt:Start()
	if not RW:InTable(plugin.runningBars, self) then tinsert(plugin.runningBars, self) end
	plugin:EnableUpdate(true)
end

function barProt:Stop()
	RW:TRemove(plugin.runningBars, self)
	RW:TRemove(self.anchor.bars, self)
	plugin:UpdateAnchor(self.anchor)
	self.bar:SetAlpha(0)
	self.bar:Hide()
	self.id = nil
	self.text = nil
	self.timeLeft = nil
	self.totalTime = nil
	self.icon = nil
	self.fadeTime = nil
	self.anchor = nil
	tinsert(plugin.freeBarObjects, self)
	if #plugin.runningBars < 1 then plugin:EnableUpdate(false) end
end

function barProt:SetText(text)
	self.bar.text:SetText(text or self.text)
end

function barProt:SetTime(time)
	self.totalTime = time
	self.timeLeft = time
	self:Update(0)
end

function barProt:SetElapsed(time)
	self.timeLeft = self.totalTime - time
end

function barProt:Update(elapsed)
	self.timeLeft = self.timeLeft - elapsed
	if self.timeLeft < 0 then
		if not self.fadeOutTime then
			self.fadeOutTime = plugin.db.profile.fadeOutTime
			self.bar.bar:SetStatusBarColor(1, 0, 0, 1)
			plugin:SendMessage("RW_BarFinished", self.id, self.text, self.icon)
		elseif self.fadeOutTime < 0.1 then
			self:Stop()
			self.fadeOutTime = nil
			return
		else
			self.fadeOutTime = self.fadeOutTime - elapsed
			self.bar:SetAlpha(self.fadeOutTime / plugin.db.profile.fadeOutTime)
		end
	else
		if self.fadeTime then
			local newAlpha = (self.totalTime - self.timeLeft) / self.fadeTime
			if newAlpha >= 0.95 then
				self.bar:SetAlpha(1)
				self.fadeTime = nil
			else
				self.bar:SetAlpha(newAlpha)
			end
		end
		self.bar.timeText:SetText(RW:TimeToStringShort(self.timeLeft))
		self.bar.bar:SetValue(1 - self.timeLeft/self.totalTime)
	end
	-- Animate movement
	if self.moveElapsed then
		if plugin.db.profile.Animate then 
			self:Animate(elapsed)
			if self.moveElapsed < 0 then
				self.moveElapsed = nil
				self:SetPosition()
			end
		else
			self:SetPosition()
			self.moveElapsed = nil
		end
	end
	plugin:FindAnchor(self)
	-- Animate Size
	if self.changeAnchor then 
		if plugin.db.profile.Animate then 
			self.changeAnchor = self.changeAnchor - elapsed
			self:AnimateSize(elapsed) 
			if self.changeAnchor < 0 then 
				self.changeAnchor = nil 
			end
		else
			self.bar:SetWidth(self.anchorSettings.Size.Width)
			self.bar:SetHeight(self.anchorSettings.Size.Height)
			self.changeAnchor = nil
		end
	end
end

function barProt:ApplySettings()
	local db = plugin.db.profile.Master
	local bar = self.bar
	
	local inset
	if db.BorderSize < 2 then
		inset = 0
	else
		inset = db.Insets
	end
	
	backdrop.bgFile = LSM:Fetch("background", db.BGTexture)
	backdrop.edgeFile = LSM:Fetch("border", db.Border)
	backdrop.edgeSize = db.BorderSize
	bar:SetBackdrop(backdrop)
	bar:SetBackdropColor(unpack(db.BGColor))
	bar:SetBackdropBorderColor(unpack(db.BorderColor))	
	
	bar.bar:SetStatusBarTexture(LSM:Fetch("statusbar", db.BarTexture))
	if self.color then
		bar.bar:SetStatusBarColor(unpack(self.color))
	else
		bar.bar:SetStatusBarColor(unpack(db.BarColor))
	end
	
	bar.timeText:SetFont(LSM:Fetch("font", db.Font), db.FontSize, db.FontOutline)
	bar.text:SetFont(LSM:Fetch("font", db.Font), db.FontSize, db.FontOutline)
	bar.text:SetJustifyH(db.FontAlign)
	bar.text:SetTextColor(unpack(db.FontColor))
	
	bar.icon:SetTexture(self.icon)
	bar.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
end

function barProt:SetPosition()
	local db = self.anchorSettings
	local anchor = self.anchor
	
	local i = self:GetOrder()
	if db.GrowUp then
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0 , db.Size.Height * i + db.Size.Spacing * i)
	else
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -anchor:GetHeight() - db.Size.Height - db.Size.Height * i - db.Size.Spacing * i)
	end
	self.bar:SetWidth(db.Size.Width)
	self.bar:SetHeight(db.Size.Height)
	
	local margin
	if plugin.db.profile.Master.BorderSize < 2 or plugin.db.profile.Master.Border == "None" then 
		margin = 0 else margin = plugin.db.profile.Master.BorderMargin 
	end
	if self.icon then
		self.bar.bar:SetWidth(self.bar:GetWidth() - self.bar:GetHeight() - margin)
	else
		self.bar.bar:SetWidth(self.bar:GetWidth() - margin * 2)
	end
end

function barProt:Animate(elapsed)
	local db = plugin.db.profile
	local barDB = self.anchorSettings
	local anchor = self.anchor
	
	local startX, startY = self.bar:GetLeft(), self.bar:GetBottom()
	local targetX, targetY = anchor:GetLeft(), 0
	local order = self:GetOrder()
	
	if barDB.GrowUp then
		targetY = anchor:GetTop() + barDB.Size.Height * order + barDB.Size.Spacing * order
	else
		targetY = anchor:GetTop() - barDB.Size.Height * order - barDB.Size.Spacing * order - anchor:GetHeight() - barDB.Size.Height
	end
	
	local newX = RW:Lerp(startX, targetX, elapsed * db.AnimSpeed)
	local newY = RW:Lerp(startY, targetY, elapsed * db.AnimSpeed)
	local diffx = -(targetX - newX)
	local diffy = -(targetY - newY)
	
	if barDB.GrowUp then
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", diffx, diffy + barDB.Size.Height * order +barDB.Size.Spacing * order)
	else
		self.bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", diffx, diffy - anchor:GetHeight() - barDB.Size.Height - barDB.Size.Height * order - barDB.Size.Spacing * order)
	end
	
	if self.changeAnchor then
		self:AnimateSize(elapsed)
	end
end

function barProt:AnimateSize(elapsed)
	local db = self.anchorSettings
	local master = plugin.db.profile.Master
	local anchor = self.anchor
	
	local newWidth = RW:Lerp(self.bar:GetWidth(), db.Size.Width, elapsed * plugin.db.profile.AnimSpeed)
	local newHeight = RW:Lerp(self.bar:GetHeight(), db.Size.Height, elapsed * plugin.db.profile.AnimSpeed)
	self.bar:SetWidth(newWidth)
	self.bar:SetHeight(newHeight)
	
	local margin
	if master.BorderSize < 2 or master.Border == "None" then margin = 0 else margin = master.BorderMargin end
	if self.icon then
		self.bar.bar:SetWidth(self.bar:GetWidth() - self.bar:GetHeight() - margin)
	else
		self.bar.bar:SetWidth(self.bar:GetWidth() - margin * 2)
	end
end

function barProt:GetOrder()
	table.sort(self.anchor.bars, barSort)
	for i, bar in pairs(self.anchor.bars) do
		if bar == self then return i-1 end
	end
end

function plugin:GetBar(id)
	for i, bar in self:BarIterator() do
		if bar.id == id then return bar end
	end
end

function plugin:FindAnchor(bar)
	local bestAnchor
	for i, anchor in pairs(self.db.profile.Anchors) do
		if bestAnchor then
			if self.db.profile.Anchors[bestAnchor].Time < bar.timeLeft and anchor.Enabled then
				if anchor.Time >= self.db.profile.Anchors[bestAnchor].Time and anchor.Enabled then
					bestAnchor = i
				end
			elseif anchor.Time < self.db.profile.Anchors[bestAnchor].Time and anchor.Time >= bar.timeLeft and anchor.Enabled then
				bestAnchor = i
			end
		elseif anchor.Enabled then
			bestAnchor = i
		end
	end
	
	if bar.anchor and bar.anchor ~= self.anchors[bestAnchor] then
		RW:TRemove(bar.anchor.bars, bar)
		self:UpdateAnchor(bar.anchor)
		bar.changeAnchor = 4
		bar.anchorSettings = self.db.profile.Anchors[bestAnchor]
		bar.anchor = self.anchors[bestAnchor]
		tinsert(bar.anchor.bars, bar)
		bar.totalTime = bar.anchorSettings.Time
		self:UpdateAnchor(bar.anchor)
	elseif not bar.anchor then
		bar.anchorSettings = self.db.profile.Anchors[bestAnchor]
		bar.anchor = self.anchors[bestAnchor]
		tinsert(bar.anchor.bars, bar)
		self:UpdateAnchor(bar.anchor)
	end
end

function plugin:UpdateAnchor(anchor)
	for i, b in pairs(anchor.bars) do
		b.moveElapsed = 4
		b.changeAnchor = 4
	end
end

function plugin:BarIterator()
	return pairs(self.runningBars)
end

----------------------
-- Message Handling --
----------------------

function plugin:RW_StartBar(message, id, time, text, icon, color)
	if not time or time <= 0 then return end
	if not self.anchors then createAnchors() end
	local newBar = self:GetBar(id)
	if newBar then 
		newBar:Stop()
		newBar = nil
	end
	
	newBar = tremove(self.freeBarObjects)
	if not newBar then
		newBar = setmetatable({
			bar = RW:CreateBar(rightClick)
		}, {__index = barProt})
	end
	
	newBar.id = id
	newBar.text = text
	newBar.timeLeft = time
	newBar.totalTime = time
	newBar.icon = icon
	newBar.color = color
	newBar.fadeTime = self.db.profile.fadeTime
	newBar.bar:Show()
	newBar.bar.obj = newBar
	
	self:FindAnchor(newBar)
	
	newBar:ApplySettings()
	newBar:SetPosition()
	newBar:SetText()
	newBar:Start()
end

function plugin:RW_CancelBar(message, id)
	local bar = self:GetBar(id)
	if bar then bar:Stop() end
end

function plugin:RW_DefaultsChanged(message, option, value)
	self.db.profile.Master[option] = value
	for i, bar in pairs(self.runningBars) do
		bar:ApplySettings()
	end
end

local configIcons = {
	[[Interface\ICONS\Ability_Hunter_DisarmingShot]],
	[[Interface\Icons\INV_Misc_EngGizmos_27]],
	[[Interface\Icons\Spell_Shadow_ScourgeBuild]],
	[[Interface\Icons\Spell_Shaman_BlessingOfTheEternals]],
	[[Interface\Icons\Spell_Shadow_Misery]],
}

function plugin:RW_ConfigMode(message, start)
	local function randomBars()
		self:RW_StartBar("2", 1, 10, L["RANDOM_MESSAGE_"..math.random(1, 4)], configIcons[math.random(1, 5)])
		self:RW_StartBar("2", 2, 15, L["RANDOM_MESSAGE_"..math.random(1, 4)], configIcons[math.random(1, 5)])
		self:RW_StartBar("2", 3, 20, L["RANDOM_MESSAGE_"..math.random(1, 4)], configIcons[math.random(1, 5)])
		self:RW_StartBar("2", 4, 45, L["RANDOM_MESSAGE_"..math.random(1, 4)], configIcons[math.random(1, 5)])
		self:RW_StartBar("2", 5, 80, L["RANDOM_MESSAGE_"..math.random(1, 4)], configIcons[math.random(1, 5)])
		self:RW_StartBar("2", 6, 95, L["RANDOM_MESSAGE_"..math.random(1, 4)], configIcons[math.random(1, 5)])
		self:RW_StartBar("2", "Restart", 30, L["RANDOM_MESSAGE_RESTART"], configIcons[math.random(1, 5)])
	end
	if start then
		if not self.anchors then createAnchors() end
		for i, anchor in pairs(self.anchors) do
			anchor:Show()
		end
		self:ScheduleRepeatingTimer(randomBars, 30)
		randomBars()
	else
		for i, anchor in pairs(self.anchors) do
			anchor:Hide()
		end
		self:CancelAllTimers()
		self:RW_CancelBar("RW_CancelBar", 1)
		self:RW_CancelBar("RW_CancelBar", 2)
		self:RW_CancelBar("RW_CancelBar", 3)
		self:RW_CancelBar("RW_CancelBar", 4)
		self:RW_CancelBar("RW_CancelBar", 5)
		self:RW_CancelBar("RW_CancelBar", 6)
		self:RW_CancelBar("RW_CancelBar", "Restart")
	end
end
-----------------------------------------------------------------

do
	local db
	
	local function insertDurationOptions()
		local option = {}
		local count = 1
		for i, setting in pairs(db.Anchors) do
			option[i.."Enabled"] = {
				type = "toggle",
				name = "Enable "..i,
				order = count,
				get = function(info) return db.Anchors[i].Enabled end,
				set = function(info, value) db.Anchors[i].Enabled = value end,
			}
			option[i.."Duration"] = {
				type = "range",
				name = i.." Bars",
				step = 1,
				min = 1,
				max = 10000,
				order = count+1,
				get = function(info) return db.Anchors[i].Time end,
				set = function(info, value) db.Anchors[i].Time = value end,
				disabled = function() return not db.Anchors[i].Enabled end
			}
			count = count + 2
		end
		return option
	end
	local function insertSizeOptions()
		local option = {}
		local count = 1
		for i, settings in pairs(db.Anchors) do
			option[i.."Width"] = {
				type = "range",
				name = i.." "..L.WIDTH,
				min = 10,
				max = 600,
				step = 1,
				order = count,
				get = function(info) return db.Anchors[i].Size.Width end,
				set = function(info, value) db.Anchors[i].Size.Width = value; self:UpdateAnchor(self.anchors[i]) end,
				disabled = function() return not db.Anchors[i].Enabled end,
			}
			option[i.."Height"] = {
				type = "range",
				name = i.." "..L.HEIGHT,
				min = 10,
				max = 600,
				step = 1,
				order = count,
				get = function(info) return db.Anchors[i].Size.Height end,
				set = function(info, value) db.Anchors[i].Size.Height = value; self:UpdateAnchor(self.anchors[i]) end,
				disabled = function() return not db.Anchors[i].Enabled end,
			}
			count = count + 2
		end
		return option
	end
	
	function plugin:GetOptions()
		db = plugin.db.profile
		local options = {
			type = "group",
			name = "Bars",
			args = {
				Animation = {
					type = "group",
					name = L.ANIMATION_SETTINGS,
					guiInline = true,
					order = 3,
					args = {
						animate = {
							type = "toggle",
							name = L.ANIMATE_TRANSITIONS,
							desc = L.ANIMATE_TRANSITIONS_DESC,
							get = function() return db.Animate end,
							set = function(info, value) db.Animate = value end,
							order = 3,
						},
						animspeed = {
							type = "range",
							name = L.ANIMATE_SPEED,
							desc = L.ANIMATE_SPEED_DESC,
							disabled = function() return not db.Animate end,
							get = function() return db.AnimSpeed end,
							set = function(info, value) db.AnimSpeed = value end,
							min = 0.1,
							max = 5,
							step = 0.1,
							order = 4
						},
					},
				},
				Durations = {
					type = "group",
					name = L.BAR_DURATION,
					guiInline = true,
					order = 5,
					args = insertDurationOptions(),
				},
				Size = {
					type = "group",
					name = L.SIZE_SETTINGS,
					guiInline = true,
					order = 6,
					args = insertSizeOptions()
				}
			}
		}
		return options
	end
end